#include <ncbi.h>
#include <gishlib.h>
#include "blastapp.h"
/* get_header(hlp) -- retrieve the header for the provided HitList */
char *
get_header(hlp)
	HitListPtr	hlp;
{
	CharPtr	cp0, cp, cpmax, cp2;
	char	ch;
	int	len;

	hlp->header_len = header_beg[hlp->seq_nbr+1] - header_beg[hlp->seq_nbr] - 1;
	hlp->header = (char *)ckalloc(hlp->header_len + 1);
	if (mfil_seek(hfile, header_beg[hlp->seq_nbr]+1, SEEK_SET) == -1)
		bfatal(ERR_MFILE, "Seek error on header file");
	if (mfil_read(hlp->header, hlp->header_len, 1, hfile) != 1)
		bfatal(ERR_MFILE, "Error reading header for sequence %ld", hlp->seq_nbr);
	hlp->header[hlp->header_len] = NULLB;

	if (gi_option) { /* suppress gi identifiers */
		cp = cp0 = hlp->header;
		len = hlp->header_len;
		for (; cp != NULL; cp = str_chr(cp+1, '\001')) {
			if (cp != cp0)
				*cp++ = '>';
			if (strncmp(cp, "gi|", 3) == 0) {
				cp2 = cp + 3;
				cpmax = cp0 + len;
				while (cp2 < cpmax) {
					ch = *cp2++;
					if (ch == '|') {
						len -= (cp2 - cp);
						mem_cpy(cp, cp2, len - (cp - cp0) + 1);
						break;
					}
					if (isspace(ch)) /* gi is the only identifier, so keep it */
						break;
				}
			}
		}
		hlp->header_len = len;
	}
	else {
		cp = hlp->header;
		while ((cp = str_chr(cp, '\001')) != NULL) {
			*cp++ = '>';
		}
	}

	return hlp->header;
}


void
get_headers(dbseqtype, hlp0)
	int	dbseqtype;
	HitListPtr	hlp0;
{
	char	*extension;
	HitListPtr	hlp;
	char	name[FILENAME_MAX];
	register HitListPtr *hlist;
	HitListPtr	*hlist0;
	size_t  i, nlists;

	if (header_beg == NULL)
		bfatal(ERR_UNDEF, "header_beg has not been initialized!");

	if (hlp0 != NULL && hfile == NULL) {
		if (dbseqtype == AMINO_ACID_SEQTYPE)
			extension = AA_HEADER_EXT;
		else
			extension = NT_HEADER_EXT;
		strcpy(name, dbfname);
		strcat(name, extension);
		hfile = mfil_open(name, "r", MFIL_OPT_ALL);
		if (hfile == NULL)
			bfatal(ERR_UNDEF, "Can't find database file ``%s''.", name);
	}


	/* Sort hitlists by seqid for faster disk seeks to get each header */

	nlists = hitlist_count(hlp0);

	/* Create a linear list for sorting */
	hlist = hlist0 = (HitListPtr *)ckalloc(sizeof(hlist[0])*nlists);
	for (hlp = hlp0; hlp != NULL; hlp = hlp->next)
		*hlist++ = hlp;

	if (numprocs > 1)
		(void) hsort((char *)hlist0, nlists, sizeof(*hlist0), (int (*)())cmp_hitlists_by_seqid);

	for (i=0; i<nlists; ++i)
		(void) get_header(hlist0[i]);

	mem_free((char *)hlist0);
}
