#include <ncbi.h>
#include <gishlib.h>
#include "blastapp.h"

#define PAGE_WIDTH	80

#define MULTI_CHR	'='
#define SINGLE_CHR	':'
#define VAXIS_CHR	'|'

void
print_hist(score_hist_dim, score_hist, S, T, lambda, K, H, qlen, dblen, nbr_seqs)
	Score_t	score_hist_dim;
	unsigned long	*score_hist;
	Score_t	S, T/* call with T==0 if there is no neighborhood to speak of */;
	double	lambda, K, H;
	Coord_t	qlen;
	unsigned long	dblen, nbr_seqs;
{
	register long	i;
	long	s;
	long	score_hist_max, score_hist_stop;
	unsigned long	observed;
	int		j, ncols, plotcols, npercol;
	Boolean		hadnonzero = FALSE;
	double	expect, old_expect, cmpexpect;
	int		swidth, cwidth;

	if (showhist == 0)
		return;

	for (i=score_hist_max=observed=0; i<=score_hist_dim; ++i) {
		observed += score_hist[i];
		if (score_hist[i] > score_hist_max)
			score_hist_max = score_hist[i];
	}

	score_hist_stop = score_hist_dim;
	for (i=score_hist_dim-1; i>=0; --i) {
		if (score_hist[i] != 0) {
			score_hist[score_hist_stop = i+1] = score_hist[score_hist_dim];
			break;
		}
	}

	swidth = longwidth10(score_hist_stop); /* width of highest score in ASCII */
	swidth = MAX(swidth, 2);
	cwidth = longwidth10(score_hist_max); /* width of highest count */
	cwidth = MAX(cwidth, 3);

	plotcols = PAGE_WIDTH - (swidth+cwidth+5);
	npercol = ceil((double)score_hist_max/(double)plotcols);
	npercol = MAX(npercol, 1); /* at least one sequence per mark */

	old_expect = HUGE_VAL;
	expect = stoe(T, lambda, K, H, dblen, (unsigned long)qlen);
	expect = MIN(1000., expect);
	expect = MIN(expect, 0.01 * nbr_seqs); /* predict extreme values only */
	cmpexpect = exp(LN10 * floor(LOG10(expect)));


	printf("\n\n    Highest-scoring Hit Extension vs. Count of Database Sequences\n\n");
	printf("        Histogram units:    %c %d Sequence%s",
				MULTI_CHR, npercol, PLURAL(npercol));
	if (npercol > 1)
		printf("     %c less than %d sequences\n", SINGLE_CHR, npercol);
	else
		putchar('\n');
	putchar('\n');

	printf(" Highest\n");
	printf(" Score\n");
	printf("%*s|  Count-->\n", swidth+1, "");
	printf("%*sV\n", swidth+1, "");

	for (i=0; i<=score_hist_stop; observed -= score_hist[i++]) {
		if (i < T && i > 0 && !hadnonzero) {
			if (score_hist[i] == 0)
				continue;
			printf(" Sub-threshold perfect-match word hits\n");
			hadnonzero = TRUE;
		}
		if (i == T)
			printf(" Neighborhood word score threshold, T = %d\n", T);
		expect = stoe((Score_t)i, lambda, K, H, dblen, (unsigned long)qlen);
		if (i == S) {
			printf(
			" >>>>>>>>>>>>>  Expect = %#0.2lg, Observed = %s  <<<<<<<<<  Cutoff Score %s\n",
			expect, lu2s(observed), lu2s(S));
			if (expect < cmpexpect && old_expect > cmpexpect)
				cmpexpect /= 10.;
		}
		else
			if (expect < cmpexpect && old_expect > cmpexpect) {
				cmpexpect /= 10.;
				printf(" -------------  Expect = %#0.2lg, Observed = %s  -----------\n",
					expect, lu2s(observed));
			}

		putchar(' ');
		if (i == score_hist_stop) {
			putchar('>');
			s = i-1;
		}
		else {
			putchar(' ');
			s = i;
		}
		printf("%*d %*d %c", swidth, s, cwidth, score_hist[i], VAXIS_CHR);
		ncols = score_hist[i]/npercol;
		for (j=0; j<ncols; ++j)
			putchar(MULTI_CHR);
		if (j == 0 && score_hist[i] > 0)
			putchar(SINGLE_CHR);
		putchar('\n');
		old_expect = expect;
	}
	putchar('\n');
}
