#include <ncbi.h>
#include <gishlib.h>
#include "blastapp.h"
#include "aabet.h"

extern struct record firstrec[RECORD_MAX], *rec;
extern char	*q_seq, *qname;
extern char	**pts;
extern char	**names;
extern long	*lens;
extern double E, H, Lambda, Lambda3, K, K3, ehigh, elow3, Efactor;
extern Score_t	S, High, Low3;
extern int showdiags, nrec;
extern int	maxnamelen;
extern Score_t	matrx[ALPHAVAL_MAX+1][ALPHAVAL_MAX+1];

static double kmn, kmn3;
static int	seqct;
static double	dmin3(), expect2(), expect3();
static int		min3(), search2(), compar(), compar3();
static int 	column[QUERYLEN_MAX];
static int	n3rec, rowct;
static struct record	*rec3, *lastrec, *frec;
static void	do3search();
static int	report_stats PROTO((FILE *));

int
out3()
{
	int		j, k, L2, L3, s1, s2, s3;
	long	maxdg;
	int		test, fcol, lcol;
	char	*S2, *S3;
	register char	*let1, *let2, *let3;
	register int	i, last;
	int		maxnl = -maxnamelen;
	CharPtr	fmt = "%*s";

	maxdg = (long)E;

	rowct=0;
	kmn = K * Meff * Neff;
	kmn3 = K3 * Meff * Neff * Neff / 2.0;
	if (High == 0)
		High = etos(ehigh, Lambda, K, H, Neff, Meff);
	if (Low3 == 0)
		Low3 = ceil(log(kmn3/elow3) / Lambda3);

	lastrec = firstrec+nrec;
	(void) hsort((char huge *)firstrec, nrec, sizeof(firstrec[0]), compar);

	for (seqct=test=0, rec=firstrec; rec<lastrec; ++rec) {
		if (showdiags) {
			printf("S:%4ld   E: %-#7.2g  Seq: ",
					(long)rec->score, expect2(rec->score));
			printf(fmt, maxnl, names[rec->seq2]);
			printf("  Start1:%4d   Start2:%4d   Length:%4d\n",
			(int)rec->start1+1, (int)rec->dg2+rec->start1+1, (int)rec->length);
		}
		if (rec->score<=High && !test) {
			test = 1;
			frec = rec;
			if (rec+maxdg < lastrec) {
				lastrec = rec+maxdg;
				while (lastrec > frec && lastrec->score == (lastrec-1)->score)
					--lastrec;
				if (lastrec == frec) {
					lastrec = rec+maxdg;
					printf("More than E=%lg diagonals had a score of %ld\n",
							E/*maxdg*/, (long)rec->score);
				}
			}
			S = (lastrec-1)->score;
		}
		if (test)
			++seqct;
	}
	putchar('\n');

	fflush(stdout);
	report_stats(stdout);

	if (frec == NULL) {
		printf("*****************************************************\n");
		printf("*\n");
		printf("*** No HSPs remain to merge into 3-way alignments\n");
		printf("*\n");
		printf("*****************************************************\n");
		return 1;
	}

	rec3=lastrec;
	n3rec = 0;

	printf("Constructing 3-alignments with score of at least %ld.\n",
			(long)Low3);

#ifdef SIGXCPU
	signal(SIGXCPU, SIG_DFL);
#endif

	RunWild("3-Way Search", (seqct*seqct - seqct - 1)/2, do3search);

#ifdef SIGXCPU
	/* Report results without interruption */
	signal(SIGXCPU, SIG_IGN);
#endif

	(void) hsort((char huge *)frec, seqct, sizeof(*frec), compar3);

	printf("\n\n");

	for (i=0; i<q_len; ++i)
		column[i]=0;
	for (rec=frec, j=0; j<rowct; ++rec, ++j) 
		for (i=rec->length-1; i>=0; --i)
			column[rec->start1+i] = 1;
	for (fcol=0; fcol<q_len; fcol=lcol) {
		while (fcol<q_len && column[fcol]==0)
			++fcol;
		if (fcol == q_len)
			break;
		for (lcol=fcol; lcol<q_len && column[lcol];)
			++lcol;
		printf(fmt, maxnl, qname);
		printf(fmt, maxnl, "");
		printf("%8s%5d ", "", fcol+1);
		for (i=fcol;i<lcol;++i)
			putchar(aa_btoa[q_seq[i]]);
		printf("%5d  S     E\n", lcol);
		for (rec=frec, j=0; j<rowct; ++rec, ++j)
			if (rec->start1 >= fcol && rec->start1 < lcol) {
				S2 = pts[rec->seq2];
				putchar('[');
				printf(fmt, maxnl, names[rec->seq3]);
				printf("%5d] ", rec->start1+rec->dg3+1);
				printf(fmt, maxnl, names[rec->seq2]);
				printf("%5d ", rec->start1+rec->dg2+1);
				for (i=fcol; i<rec->start1; ++i)
					putchar(' ');
				for (k=0; k<rec->length; ++k)
					fputc(aa_btoa[S2[rec->dg2+i++]], stdout);
				for (; i<lcol; ++i)
					putchar(' ');
				printf("%5d%4d %-#7.2lg\n",
							rec->start1+rec->length+rec->dg2,
							rec->scor3,
							expect3(rec->scor3));
			}
		putchar('\n');
	}

	(void) hsort((char huge *)lastrec,n3rec,sizeof(struct record),compar);

	for (rec=lastrec; rec<rec3; ++rec) {
		S2 = pts[rec->seq2];
		L2 = lens[rec->seq2];
		S3 = pts[rec->seq3];
		L3 = lens[rec->seq3];
		printf("3-score:%4d   E: %-#7.2lg\n", rec->score, expect3(rec->score));
		s1 = search2(q_seq, S2, q_len, L2, rec->dg2);
		s2 = search2(q_seq, S3, q_len, L3, rec->dg3);
		s3 = search2(S2, S3, L2, L3, rec->dg3-rec->dg2);
		printf("E12: %-#7.2lg  ", expect2(s1));
		printf("E13: %-#7.2lg  ", expect2(s2));
		printf("E23: %-#7.2lg\n", expect2(s3));
		printf("Diag12:%4d   ", s1);
		printf("Diag13:%4d   ", s2);
		printf("Diag23:%4d\n", s3);
		s1=s2=s3=0;
		let1=q_seq+rec->start1;
		let2=S2+rec->start1+rec->dg2;
		let3=S3+rec->start1+rec->dg3;
		last=rec->start1+rec->length;
		for (i=rec->start1; i<last; ++i) {
			s1 += matrx[*let1][*let2];
			s2 += matrx[*let1++][*let3];
			s3 += matrx[*let2++][*let3++];
		}
		printf("Scor12:%4d   Scor13:%4d   Scor23:%4d\n\n",s1,s2,s3);
		printf(fmt, maxnl, qname);
		printf("%5d ", rec->start1+1);
		for (i=rec->start1; i<last; ++i)
			putchar(aa_btoa[q_seq[i]]);
		printf(" %d\n", last);
		printf(fmt, maxnl, names[rec->seq2]);
		printf("%5d ", rec->start1+1+rec->dg2);
		for (i=rec->start1; i<last; ++i)
			putchar(aa_btoa[S2[i+rec->dg2]]);
		printf(" %d\n", last+rec->dg2);
		printf(fmt, maxnl, names[rec->seq3]);
		printf("%5d ", rec->start1+1+rec->dg3);
		for (i=rec->start1; i<last; ++i)
			putchar(aa_btoa[S3[i+rec->dg3]]);
		printf(" %d\n\n",last+rec->dg3);
	}
	return 0;
}

static void
do3search(userp, tp)
	Nlm_VoidPtr	userp;
	TaskBlkPtr	tp;
{
	int		k, L2, L3, off2, off3;
	int		mx;
	double	e23;
	char	*S2, *S3;
	register char	*let1, *let2, *let3;
	register int	i, f, begin, end, sum, maxsum, first = 0, last = 0;
	struct record	*rec, *rec2;

	if (tp->id == 0)
		signal(SIGTERM, SigTerm);

	for (k = 0, rec = frec+tp->id; rec<lastrec; rec += tp->proc_max) {
		S3 = pts[rec->seq2];
		L3 = lens[rec->seq2];
		off3 = rec->dg2;
		for (rec2 = rec-1; rec2 >= frec; --rec2) {
			TaskTickNext(tp, k += tp->proc_max);
			S2 = pts[rec2->seq2];
			L2 = lens[rec2->seq2];
			off2 = rec2->dg2;
			mx = search2(S2,S3,L2,L3,off3-off2);
			if (mx <= High && mx+rec->score+rec2->score >= Low3) {
				begin = -min3(0, off2, off3);
				end = min3(q_len, L2-off2, L3-off3);
				maxsum = sum = 0;
				let1 = q_seq+begin;
				let2 = S2+begin+off2;
				let3 = S3+begin+off3;
				for (f = i = begin; i < end; ++i) {
					sum += matrx[*let1][*let2];
					sum += matrx[*let1++][*let3];
					sum += matrx[*let2++][*let3++];
					if (sum >= maxsum) {
						maxsum = sum;
						last = i;
						first = f;
					}
					else
						if (sum<0) {
							f = i+1;
							sum = 0;
						}
				}
				if (maxsum >= Low3 &&
						expect3(maxsum) <= Efactor*expect2(MAX(rec2->score, mx))) {
					M_LOCK;
					if (rec3 >= firstrec+RECORD_MAX)
						bfatal(ERR_RECORD3, "Out of storage for 3-way alignments; increase RECORD_MAX and recompile.");
					rec3->score = maxsum;
					rec3->seq2 = rec2->seq2;
					rec3->seq3 = rec->seq2;
					rec3->start1 = first;
					rec3->dg2 = off2;
					rec3->dg3 = off3;
					rec3->length = last-first+1;
					e23 = expect2(mx);
					if (expect3(rec3->score) <= dmin3(e23,expect2(rec->score),expect2(rec2->score))) {
						if (rec3->score > rec->scor3) {
							if (rec->scor3 == 0)
								++rowct;
							rec->scor3 = rec3->score;
							rec->seq3 = rec3->seq2;
							rec->dg3 = rec3->dg2;
							rec->start1 = rec3->start1;
							rec->length = rec3->length;
						}
						if (rec3->score > rec2->scor3) {
							if (rec2->scor3 == 0)
								++rowct;
							rec2->scor3 = rec3->score;
							rec2->seq3 = rec3->seq3;
							rec2->dg3 = rec3->dg3;
							rec2->start1 = rec3->start1;
							rec2->length = rec3->length;
						}
					}
					++rec3;
					++n3rec;
					M_UNLOCK;
				}
			}
		}
	}
}

static double
dmin3(a, b, c)
	register double	a,b,c;
{
	if (a < b)
		return MIN(a,c);
	return MIN(b,c);
}

static int
min3(a, b, c)
	register int	a,b,c;
{
	if (a < b)
		return MIN(a,c);
	return MIN(b,c);
}

static int
search2(seq1, seq2, len1, len2, dg)
	char	*seq1, *seq2;
	int		len1, len2, dg;
{
	register int	sum, maxsum;
	register char	*let1, *let2, *maxlet1;
	register int	pos1, pos2;

	if (dg > -len1 && dg < len2) {
		pos1 = MAX(0, -dg);
		pos2 = dg+pos1;
		let1 = seq1+pos1;
		let2 = seq2+pos2;
		maxlet1 = let1 + MIN(len1-pos1, len2-pos2);
		maxsum = sum = 0;
		while (let1 < maxlet1)
			if ((sum += matrx[*let1++][*let2++]) < 0)
				sum = 0;
			else
				if (sum > maxsum)
					maxsum = sum;
		return maxsum;
	}
	return 0;
}

static double
expect2(a)
	int 	a;
{
	return kmn*exp(-Lambda*a);
}

static double
expect3(a)
	int		a;
{
	return kmn3*exp(-Lambda3*a);
}

static int
compar(s1, s2)
	register struct record *s1, *s2;
{
	register int	diff;

	diff = s2->score - s1->score;
	if (diff != 0)
		return diff;
	diff = s1->length - s2->length;
	if (diff != 0)
		return diff;
	return s1->seq2 - s2->seq2;
}

static int
compar3(s1,s2)
	register struct record *s1, *s2;
{
	register int	diff;

	diff = s2->scor3 - s1->scor3;
	if (diff != 0)
		return diff;

	diff = s2->length - s1->length;
	if (diff != 0)
		return diff;

	diff = s1->start1 - s2->start1;
	if (diff != 0)
		return diff;

	diff = s1->seq2 - s2->seq2;
	if (diff != 0)
		return diff;

	return s1->seq3 - s2->seq3;
}


static int
report_stats(fp)
	FILE	*fp;
{
	fprintf(fp, "Query length: %s   Database length: %s\n\n",
			lu2s(Meff), lu2s(Neff));
	fprintf(fp, "For 2-alignments K = %#0.3lg and Lambda = %#0.3lg   ",
			K, Lambda);
	fprintf(fp, "Expect( S >%5ld ) = %#0.2lg\n", (long)High, expect2(High+1));
	fprintf(fp, "For 3-alignments K = %#0.3lg and Lambda = %#0.3lg   ",
			K3, Lambda3);
	fprintf(fp, "Expect( S >=%4ld ) = %#0.2lg\n\n", (long)Low3, expect3(Low3));
	fprintf(fp, "Number of diags with score between %ld and %ld inclusive:  %d.\n",
				(long)S, (long)High, seqct);
	return 0;
}
