
static void	extend_aa PROTO((MPPROTOC CharPtr s, CharPtr q));

/* -----------------------   search a database sequence   --------------------*/
static void
search_aa(MP)
	MPDEF
{
	register DFA_StatePtr state;
	register unsigned char *s;
	register DFA_PatlistPtr	plp;
	unsigned	hspcnt;

	hspcnt = MPPTR curhl.hspcnt;

	s = (unsigned char *)MPPTR s_seq;
	if (dfa_check(dp) != dfaErrNone)
		bfatal(ERR_DFALIB, "corrupted DFA pointer; Could not continue");
Restart:
	state = dp->state0;

	for (;;) {
		/*
			The machine "accepts" when a word hit is encountered,
			and "halts" on the null terminator of s or any other
			non-alphabetic character encountered in s.
		*/
		while (!DFA_ISACCEPTING(state = state->next[*s++]))
			;

		/* Has finite-state machine halted (e.g., end of sequence)? */
		if ((state = (DFA_StatePtr)DFA_UNMUNGED(state)) == NULL) {
			/* Hit a gap character or a sentinel null terminator? */
			if ((s - (unsigned char *)MPPTR s_seq) >= MPPTR s_len)
				break;
			goto Restart;
		}

		/* Extend each hit in the linked list */
		plp = &((DFA_AcceptPtr)state)->pl;
		state = ((DFA_AcceptPtr)state)->retState;

		do {
			extend_aa(MPC (CharPtr)s, plp->patID.p);
			/* unroll one iteration of the loop */
			if ((plp = plp->chain) == NULL)
				break;
			extend_aa(MPC (CharPtr)s, plp->patID.p);
		} while ((plp = plp->chain) != NULL);
	}

    /* Clear the record of all diagonal hit positions */
    {
        register DIAG_PTR   dt = MPPTR diag_top;
        register DIAG_PTR   ds = MPPTR diag_stack;
        register DIAG_TYPE  *dl = MPPTR diag_level;

        while (dt > ds)
            dl[*--dt] = 0;

        MPPTR diag_top = ds;
    }
}

/* extend_aa - extend a word-sized hit to a longer match */
static void
extend_aa(MPC s_pos, q_pos)
	MPDEF
	register CharPtr	s_pos;
	register CharPtr	q_pos;
{
	register char *q, *s;
	register Score_t	sum;
	register Score_t	x;
	register Score_t	score;
	register char *q_beg, *q_end;
	register Coord_t	diag;
	CharPtr	qf, sf;

	MPPTR numhits++;
	x = q_pos - q_seq;
	if (MPPTR diag_level[diag = (s_pos - MPPTR s_seq) - x] > x)
		return;
	if (MPPTR diag_level[diag] == 0)
		*MPPTR diag_top++ = diag;

    /*
    Find where positive scoring starts and ends within the word hit.

    This examination, which was not performed in early versions of the
	software, is necessary because some word hits may begin with
    negative-scoring letter substitution(s) which corrupt the interpretation
	of the end-point of an alignment as well as the distinction of where the
	extension score has fallen off by the quantity X or more.
    */
	q_beg = q = q_pos - W;
#ifdef TBLASTN
	q_end = NULL;
#endif
	s = s_pos - W;
	score = sum = 0;
	do
		if ((sum += matrx[*q++][*s++]) > score) {
			/*
			Except in TBLASTN...
			This block will be executed at least once, because hits only
			occur when the word score is positive
			*/
			score = sum;
			q_end = q;
		}
		else
			if (sum <= 0) {
				sum = 0;
				q_beg = q;
			}
	while (q < q_pos);

#ifdef TBLASTN
	if (q_end == NULL)
		return;
#endif

	if ((x = -score) < minusX)
		x = minusX;

	/* extend to the left */
	s = s_pos - (q_pos - (q = q_beg));
	sum = 0;
	do
		if ((sum += matrx[*--q][*--s]) > 0) {
			q_beg = q;
			score += sum;
			/* unroll one iteration of the do-while loop */
			if ((sum = matrx[*--q][*--s]) > 0) {
				q_beg = q;
				score += sum;
				sum = 0;
			}
			if ((x = -score) < minusX)
				x = minusX;
		}
	while (sum >= x); /* fails on null prefix of either sequence */

	/* extend to the right */
	s = s_pos + ((q = q_end) - q_pos);
	sum = 0;
	do
		if ((sum += matrx[*q++][*s++]) > 0) {
			q_end = q;
			score += sum;
			/* unroll one iteration of the do-while loop */
			if ((sum = matrx[*q++][*s++]) > 0) {
				q_end = q;
				score += sum;
				sum = 0;
			}
			if ((x = -score) < minusX)
				x = minusX;
		}
	while (sum >= x); /* fails on null terminator of either sequence */

	if (score < S2) {	/* Score is not reportable */
		/* Record how far this diagonal has been traversed */
		MPPTR diag_level[diag] = (q - q_seq) + W;
		++MPPTR numfailed;
		if (MPPTR high_score < score)
			MPPTR high_score = score;
		return;
	}

#ifdef TBLASTN
	/*
		Check for ambiguity codes in the underlying nucleic acid sequence
	*/
	if ((ambiguity_ray[MPPTR current/CHAR_BIT] & (1<<(MPPTR current%CHAR_BIT))) && sfile != NULL) {
		CharPtr	qmax;
		unsigned long	alen, nlen, npos, spos;
		unsigned char	qsave;
		int		failed = 1;
		size_t	needed;
		CharPtr	cp0;
		char	stakbuf[5*KBYTE];

		MPPTR diag_level[diag] = (q - q_seq) + W;
		spos = (q_beg - q_seq) + diag;
		alen = q_end - q_beg;
		nlen = alen * CODON_LEN;
		needed = nlen + nlen/line_len + 2;
		if (MPPTR frame > 0)
			npos = spos*CODON_LEN + MPPTR frame - 1;
		else
			npos =  MPPTR d_len - spos*CODON_LEN + MPPTR frame - nlen + 1;
		if (needed < DIM(stakbuf))
			cp0 = stakbuf;
		else
			cp0 = ckalloc(needed);
		M_LOCK;
		get_seq_from_fasta(cp0, npos, nlen, sfile, seq_beg[MPPTR current], line_len);
		M_UNLOCK;
		if (MPPTR frame > 0)
			trans2(cp0, MPPTR s_seq + spos, alen);
		else
			rtrans2(cp0, MPPTR s_seq + spos, alen);

		score = sum = 0;
		s = MPPTR s_seq + spos;
		/* install a temporary sentinel byte at end of query segment */
		qmax = (q = q_beg) + alen;
		qsave = *qmax;
		*qmax = aa_atob['-'];
		for (;;) {
			if ((sum += matrx[*q++][*s++]) <= 0) {
				if (score > 0) {
					q = q_end;
					s = MPPTR s_seq + (q_end - q_seq);
					if (MPPTR high_score < score)
						MPPTR high_score = score;
					if (score >= S2) {
						struct hsp	h;

						failed = 0;
						h.frame = MPPTR frame;
						h.score = score;
						h.len = q_end - q_beg;
						h.q_pos = q_beg - q_seq;
						h.s_pos = h.q_pos + diag;
						ckoverlap(MPC &h);
					}
				}
				score = sum = 0;
				q_beg = q_end = q;
				if (q >= qmax)
					break;
			}
			else
				if (sum > score) {
					score = sum;
					q_end = q;
				}
		}
		*qmax = qsave; /* restore original letter at end of query segment */
		MPPTR numfailed += failed;
		if (cp0 == stakbuf)
			return;
		free(cp0);
		return;
	}

	if (score >= S2) {
		struct hsp	h;

		if (MPPTR high_score < score)
			MPPTR high_score = score;

		/* Record the HSP */
		h.score = score;
		h.len = q_end - q_beg;
		h.s_pos = (h.q_pos = q_beg - q_seq) + diag;
		h.frame = MPPTR frame;
		ckoverlap(MPC &h);
	}

#else /* !TBLASTN */

	/*********************************************************************
	Score is reportable.  Now examine the diagonal rigorously, ignoring X.
	This was not done by the program described by Altschul et al. (1990)
	J. Mol. Biol.
	*********************************************************************/

	/* Mark this diagonal (in advance) as having been fully examined */
	MPPTR diag_level[diag] = INT_MAX;

	/* Start at the very beginning of the diagonal */
	s = MPPTR s_seq;
	q = q_seq;
	if (diag > 0)
		s += diag;
	else
		q -= diag;

	qf = q_seq + q_len;
	sf = MPPTR s_seq + MPPTR s_len;
	if (qf - q > sf - s)
		qf = q + (sf - s);
	q_beg = q;
	score = sum = 0;
	for (;;) {
		if ((sum += matrx[*q++][*s++]) <= 0) {
			if (score > 0) {
				q = q_end;
				s = MPPTR s_seq + (q_end - q_seq) + diag;
				if (score >= S2) {
					struct hsp	h;

					if (MPPTR high_score < score)
						MPPTR high_score = score;

					/* Record the HSP */
					h.score = score;
					h.len = q_end - q_beg;
					h.q_pos = q_beg - q_seq;
					h.s_pos = h.q_pos + diag;
					h.frame = MPPTR frame;
					ckoverlap(MPC &h);
				}
			}
			if (q > qf)
				break; /* Hit a sentinel byte */
			q_beg = q_end = q;
			score = sum = 0;
		}
		else
			if (sum > score) {
				score = sum;
				q_end = q;
			}
	}
#endif /* !TBLASTN */
}
