static void	extend_nt PROTO((MPPROTOC PACK_TYPE *c, Coord_t q_pos, int left, int right, char *qq_seq, int strand));

/* ----------------------- scan a nucleotide database --------------------*/
static void
search_nt(MP)
	MPDEF
{
	register DFA_StatePtr	state;
	register PACK_TYPE	*c = MPPTR c_seq - 1;
	register DFA_PatlistPtr	plp;
	register DFA_AcceptPtr	ap;
	int		strand;
	Coord_t	q_pos;
	char	*q;
	PACK_TYPE	p;
	int	left, right;
	PACK_TYPE	*end = MPPTR c_seq + MPPTR c_len;
	char	*qq_seq;

	if (dfa_check(dp) != dfaErrNone)
		bfatal(ERR_DFALIB, "corrupted DFA pointer; Could not continue");
	state = dp->state0;

	for (;;) {
		while (!DFA_ISACCEPTING(state = state->next[*++c]))
			;

		ap = (DFA_AcceptPtr)DFA_UNMUNGED(state);
		state = ap->retState;
	
		plp = &ap->pl;
		do {
			if ((q_pos = (Coord_t)plp->patID.u) > 0) {
				q = qq_for;
				strand = 1;
				/* the location of this test is the reason MAGIC_NUMBER must
				be a positive integer */
				if (q_pos == MAGIC_NUMBER) {
					if (c >= end) /* end of database? */
						return;
					continue; /* false alarm */
				}
			}
			else {
				q = qq_rev;
				q_pos = -q_pos;
				strand = -1;
			}

			++MPPTR numhits8;

			/* how many positions left does the match extend? */
			qq_seq = q;
			q_pos -= 2;
			p = c[-WORDSIZE_MIN/(CHAR_BIT/NBPN)];
			if (BASE3(p) != *(q += q_pos))
				left = 0;
			else
				if (BASE2(p) != *--q)
					left = 1;
				else
					if (BASE1(p) != *--q)
						left = 2;
					else
						if (BASE0(p) != *--q)
							left = 3;
						else
							left = 4;
		
			/* how many positions right does the match extend? */
			q = qq_seq + q_pos + (WORDSIZE_MIN+1);
			p = c[1];
			if (BASE0(p) != *q)
				right = 0;
			else
				if (BASE1(p) != *++q)
					right = 1;
				else
					if (BASE2(p) != *++q)
						right = 2;
					else
						if (BASE3(p) != *++q)
							right = 3;
						else
							right = 4;

			if (left + right >= WmWM)
				extend_nt(MPC c, q_pos+1, left, right, qq_seq, strand);

		} while ((plp = plp->chain) != NULL);
	} /* for (;;) */
}

/* extend_nt - extend word-sized hit to longer match */
static void
extend_nt(MPC c, q_pos, left, right, qq_seq, strand)
	MPDEF
	PACK_TYPE	*c;
	Coord_t	q_pos;
	register int	left, right;
	char	*qq_seq;
	int		strand;
{
	register char	*q, *d, *dlim;
	register Score_t	sum, x, n = N, m = M;
	Score_t	score;
	DIAG_TYPE	*diagentry;
	PACK_TYPE	*cbp, *clastp;
	char	*q_beg, *q_end;
	unsigned long	s_pos, len;
	Seqid_t	cur;
	register Coord_t	limit;
	long	cbyte;

	/* check whether ends of query sequence were over-run */
	if (left > q_pos)
		if (q_pos + right < WmWM)
			return;
		else
			left = q_pos;
	if ((limit = q_lenmwm - q_pos) < (int)right)
		if ((int)left + limit < WmWM)
			return;
		else
			right = limit;

	/* return if W-base hit crosses a database sequence boundary */
	if ((cbyte = c - MPPTR c_seqp1) == 0 && right < WmWM)
		return;
	if (cbyte > MPPTR c_last - 3) {
		if (MPPTR c_pad != 0) {
			if (cbyte >= MPPTR c_last-1 || (left + MIN(right, MPPTR c_pad)) < WmWM)
				return;
		}
		else {
			if (cbyte == MPPTR c_last || (cbyte == MPPTR c_last-1 && left < WmWM))
				return;
		}
	}

	MPPTR numhits++;

	{
		register Coord_t	diag;
		register DIAG_TYPE	*dl, *dt;

		/* plus-strand hits reside on even-numbered diagonals */
		/* minus-strand hits reside on odd-numbered diagonals */
		diag = cbyte*(CHAR_BIT/NBPN) - q_pos;
		diag += diag + (strand < 0);
		dl = MPPTR diag_level;
		dt = &MPPTR diag_level[MPPTR diag_top];
		for (; dl < dt; ++dl) {
			if (dl->diag == diag)
				if (dl->pos <= q_pos)
					goto GotDiagEntry;
				else
					return; /* excluded hit */
		}
		if (++MPPTR diag_top > DIAGCNT_MAX)
			bfatal(ERR_DIAGCNT, "too many matches; raise DIAGCNT_MAX and recompile, or query with a shorter sequence");
		dl->diag = diag;
GotDiagEntry:
		diagentry = dl;
	}

	/* extend to the left */
	d = dlim = NULL;
	q_beg = q = qq_seq + q_pos;
	score = MtWM;
	if ((x = -score) < minusX)
		x = minusX;
	sum = 0;
	cbp = c - 1;
	clastp = MPPTR c_lastp;
	/* extend to the left */
	while (q > qq_seq) {
		if (d == dlim) {
			if (cbp <= MPPTR c_seq)
				break;
			d = (dlim = &xray[*--cbp][0]) + DIM(xray[0]);
		}
		if ((sum += ntmatrx[*--q][*--d]) <= 0) {
			if (sum < x)
				break;
			continue;
		}
		q_beg = q;
		score += sum;
		sum = 0;
		if ((x = -score) < minusX)
			x = minusX;
	}

	/* extend to the right */
	cbp = c;
	d = dlim;
	q_end = q = qq_seq + q_pos + WORDSIZE_MIN;
	sum = 0;
	while (q < qq_seq + q_len) {
		if (d == dlim) {
			if (cbp >= clastp)
				break;
			dlim = (d = &xray[*++cbp][0]) + DIM(xray[0]);
			if (cbp == clastp && MPPTR c_pad != 0)
				dlim = d + MPPTR c_pad;
		}
		if ((sum += ntmatrx[*q++][*d++]) <= 0) {
			if (sum < x)
				break;
			continue;
		}
		q_end = q;
		score += sum;
		sum = 0;
		if ((x = -score) < minusX)
			x = minusX;
	}
	diagentry->pos = q_end - qq_seq;
	if (score < S) {
		++MPPTR numfailed;
		if (score > MPPTR high_score)
			MPPTR high_score = score;
		return;
	}

	s_pos = cbyte*(CHAR_BIT/NBPN) - (q_pos - (q_beg-qq_seq));
	len = q_end - q_beg;
	cur = MPPTR current;

	if ((ambiguity_ray[cur/CHAR_BIT] & (1<<(cur%CHAR_BIT))) && sfile != NULL) {
		register CharPtr	cp, cpmax;
		CharPtr	q_beg0;
		int		failed = 1;
		int		needed;
		CharPtr	cp0;
		char	stakbuf[5*KBYTE];

		needed = len + len/line_len + 2;
		if (needed < DIM(stakbuf))
			cp0 = stakbuf;
		else
			cp0 = ckalloc(needed);
		M_LOCK;
		get_seq_from_fasta_xl(cp0, s_pos, len, sfile, seq_beg[cur], line_len, nt_atob);
		M_UNLOCK;

		cpmax = (cp = cp0) + len;
		*cpmax = nt_atob['-']; /* install a sentinel byte */
		score = sum = 0;
		q = q_beg0 = q_beg;
		for (;;) {
			if ((sum += ntmatrx[*q++][*cp++]) <= 0) {
				if (score > 0) {
					if (MPPTR high_score < score)
						MPPTR high_score = score;
					q = q_end;
					cp = cp0 + (q_end - q_beg0);
					if (score >= S) {
						struct hsp	h;

						failed = 0;
						/* Record the HSP */
						h.score = score;
						h.len = q_end - q_beg;
						h.q_pos = q_beg - qq_seq;
						h.s_pos = s_pos + (q_beg - q_beg0);
						h.frame = strand;
						ckoverlap(MPC &h);
					}
				}
				score = sum = 0;
				q_beg = q_end = q;
				if (cp >= cpmax)
					break;
			}
			else
				if (sum > score) {
					score = sum;
					q_end = q;
				}
		}
		MPPTR numfailed += failed;
		if (cp0 == stakbuf)
			return;
		free(cp0);
		return;
	}

	{
		struct hsp	h;

		if (score > MPPTR high_score)
			MPPTR high_score = score;
		h.score = score;
		h.len = len;
		h.q_pos = q_beg - qq_seq;
		h.frame = strand;
		h.s_pos = s_pos;
		ckoverlap(MPC &h);
	}
}
