static void	extend_x PROTO((MPPROTOC char *s, Coord_t pos));

/* -----------------------   search a database sequence   --------------------*/
static void
search_x(MP)
	MPDEF
{
	register DFA_StatePtr	state;
	register char	*s;
	register DFA_PatlistPtr	plp;
	register DFA_AcceptPtr	ap;

	s = MPPTR s_seq;
	if (dfa_check(dp) != dfaErrNone)
		bfatal(ERR_DFALIB, "corrupted DFA pointer; Could not continue");
Restart:
	state = dp->state0;

	for (;;) {
		/*
			Machine "Accepts" when a word hit is encountered,
			and "Halts" on the null terminator of s or any other
			non-alphabetic character encountered in s.
		*/
		while (!DFA_ISACCEPTING(state = state->next[*s++]))
			;

		/* Has finite-state machine halted (e.g., end of sequence)? */
		if ((ap = (DFA_AcceptPtr)DFA_UNMUNGED(state)) == NULL) {
			/* Hit a gap character or null terminator? */
			if (s - MPPTR s_seq >= MPPTR s_len)
				break;
			goto Restart;
		}

		state = ap->retState;

		/* Extend each hit in the linked list */
		plp = &ap->pl;
		do {
			extend_x(MPC s, (Coord_t)plp->patID.u);
			/* unroll one iteration of the loop */
			if ((plp = plp->chain) == NULL)
				break;
			extend_x(MPC s, (Coord_t)plp->patID.u);
		} while ((plp = plp->chain) != NULL);
	}

	{
		register DIAG_TYPE	*diag_level;
		register int	*diag_top, *diag_stack;
		register int	frame;
		seq_data	*sdp;

		for (frame=minframe; frame<maxframe; ++frame) {
			sdp = &MPPTR sd[frame];
			diag_level = sdp->diag_level;
			diag_top = sdp->diag_top;
			diag_stack = sdp->diag_stack;
			while (diag_top > diag_stack)
				diag_level[*--diag_top] = 0;
			sdp->diag_top = diag_top;
		}
	}
}

/* extend_x - extend a word-sized hit to a longer match */
static void
extend_x(MPC s_pos, pos)
	MPDEF
	register char	*s_pos;
	register Coord_t	pos;
{
	register char	*q, *s;
	register Score_t	sum, x;
	register Score_t	score;
	register char	*q_pos;
	register char	*q_beg, *q_end;
	register seq_data	*sdp;
	Score_t	minusX;
	Coord_t	diag;
	char	*q_seq;

	++MPPTR numhits;

	sdp = &MPPTR sd[pos & 7];
	pos >>= 3;
	if (sdp->diag_level[diag = (s_pos - MPPTR s_seq) - pos] > pos)
		return;
	if (sdp->diag_level[diag] == 0)
		*(sdp->diag_top++) = diag;

	q_pos = (q_seq = sdp->seq) + pos;

	/* Find where positive scoring starts & ends within the word hit */
	q_beg = q = q_pos - W;
	s = s_pos - W;
	score = sum = 0;
	do
		if ((sum += matrx[*q++][*s++]) > score) {
			/*
			This block must be executed at least once, because hits only
			occur when the word score is positive.
			*/
			score = sum;
			q_end = q;
		}
		else
			if (sum <= 0) {
				sum = 0;
				q_beg = q;
			}
	while (q < q_pos);

	if ((x = -score) < (minusX = sdp->minusX))
		x = minusX;

	/* extend to the left */
	s = s_pos - (q_pos - (q = q_beg));
	sum = 0;
	do
		if ((sum += matrx[*--q][*--s]) > 0) {
			q_beg = q;
			score += sum;
			/* unroll one iteration of the do-while loop */
			if ((sum = matrx[*--q][*--s]) > 0) {
				q_beg = q;
				score += sum;
				sum = 0;
			}   
			if ((x = -score) < minusX)
				x = minusX;
		}       
	while (sum >= x); /* fails on null prefix of either sequence */
             
	/* extend to the right */
	s = s_pos + ((q = q_end) - q_pos);
	sum = 0; 
	do      
		if ((sum += matrx[*q++][*s++]) > 0) {
			q_end = q;
			score += sum;
			/* unroll one iteration of the do-while loop */
			if ((sum = matrx[*q++][*s++]) > 0) {
				q_end = q;
				score += sum;
				sum = 0;      
			}            
			if ((x = -score) < minusX)
				x = minusX;            
		}                   
	while (sum >= x); /* fails on null terminator of either sequence */
 
	sdp->diag_level[diag] = (q - q_seq) + W;
	++MPPTR numfailed; /* assume the likely--that it failed */

	if (score < sdp->S0) /* Score is not reportable */
		return;

	if (score < sdp->S) { /* Score is not reportable */
		if (score > sdp->highs_obs)
			sdp->highs_obs = score;
		return;
	}

	/* Score is reportable.  Now examine the diagonal rigorously ignoring X */
	/* (This was not done by the program described in the JMB paper) */

	/* Mark this diagonal in advance as having been fully examined */
	sdp->diag_level[diag] = (q_end - q_seq) + W;
	--MPPTR numfailed;

	/* Start at the very beginning of the diagonal */
	s = MPPTR s_seq;
	q = q_seq;
	if (diag > 0)
		s += diag;
	else
		q -= diag;

	q_beg = q;
	score = sum = 0;
	for (;;) {
		Score_t	S = sdp->S;

		if ((sum += matrx[*q++][*s++]) <= 0) {
			if (score > 0) {
				q = q_end;
				s = MPPTR s_seq + (q_end - q_seq) + diag;
				if (score >= S) {
					struct hsp	h;

					if (score > sdp->highs_obs)
						sdp->highs_obs = score;

					/* Record the HSP */
					h.score = score;
					h.len = q_end - q_beg;
					h.s_pos = (h.q_pos = q_beg - q_seq) + diag;
					h.frame = sdp->stdframe;
					ckoverlap(MPC &h);
				}
			}
			else
				if (sum < SCORE_MIN/WORDSIZE_MAX/2) {
					if (q - q_seq > sdp->len || s - MPPTR s_seq > MPPTR s_len)
						break; /* Hit a sentinel byte */
				}
			q_beg = q_end = q;
			score = sum = 0;
		}
		else
			if (sum > score) {
				score = sum;
				q_end = q;
			}
	}
	return;
}
