#include <ncbi.h>
#include <gish.h>
#include <gishlib.h>
#if defined(YESMFILE) && defined(MMAP_AVAIL)
#include <sys/mman.h>
#ifndef MAP_FILE
#define MAP_FILE	0
#endif
#ifndef MAP_AUTOGROW
#define MAP_AUTOGROW	0
#endif
#endif /* YESMFILE && MMAP_AVAIL */


static int	maplen = MFIL_MAPLEN_DEFAULT;

int _cdecl
mfil_maplen(newlen)
	int	newlen;
{
	int		oldlen;

	oldlen = maplen;
	if (newlen <= 0)
		maplen = MFIL_MAPLEN_DEFAULT;
	else
		maplen = newlen;

	return oldlen;
}

int _cdecl
mfil_inmemory(mfp)
	MFILE	PNTR mfp;
{
#ifdef YESMFILE
	return mfp->sp != NULL && mfp->shmdp != NULL;
#else
	return 0;
#endif
}

/*
	mfil_prefetch

	pre-read the file into memory, forcing system to assign VM pages
*/
void _cdecl
mfil_prefetch(mfp)
	MFILE	PNTR mfp;
{
#if defined(YESMFILE) && defined(MPROC_AVAIL)
	register CharPtr	cp, cpmax;
	register int	ch;

	if (mfp->sp == NULL)
		return;

	for (cp = mfp->sp, cpmax = mfp->sp + mfp->filesize; cp < cpmax; cp += 512)
		ch += *cp;
#else
	return;
#endif
}

/*
	mfil_get() -- return a pointer to an area of storage of length 'len'
	that is populated with data from the current MFILE pointer for length
	'len'.  The current MFILE pointer is then advanced by 'len'.

	If MFILE is not in shared memory, then the memory area is allocated
	with malloc() and the data is read into that memory from disk.

	WARNING:  data alignment problems (bus errors) may result from improper
	dereferencing of the pointers returned by this function!
*/
CharPtr _cdecl
mfil_get(mfp, len)
	MFILE	PNTR mfp;
	size_t	len;
{
	register CharPtr	cp;
#ifdef YESMFILE
	register size_t	off;

	if (mfp->sp != NULL) {
		off = mfp->offset;
		if (off + len > mfp->filesize)
			/*len = mfp->filesize - off;*/
			return NULL;
		mfp->offset += len;
		return mfp->sp + off;
	}
#endif

	cp = (CharPtr)mem_malloc(len);
	if (cp == NULL)
		return NULL;
#ifdef YESMFILE
	if (fread(cp, len, 1, mfp->fp) != 1) {
#else
	if (fread(cp, len, 1, mfp) != 1) {
#endif
		mem_free(cp);
		return NULL;
	}
	return cp;
}

size_t
mfil_put_long(mfp, lp, nelem)
	MFILE	PNTR mfp;
	long	PNTR lp;
	size_t	nelem;
{
	register long	PNTR lpmax;
	long	l;

#if BYTE_ORDER == BIG_ENDIAN
	if (sizeof(long) == BO_LONG_SIZE)
		return mfil_write((CharPtr)lp, nelem, BO_LONG_SIZE, mfp);
	lpmax = lp + nelem;
	nelem = 0;
	while (lp < lpmax) {
		nelem += mfil_write(((CharPtr)lp)+(sizeof(long)-BO_LONG_SIZE), 1, BO_LONG_SIZE, mfp);
		++lp;
	}
	return nelem;
#else
	if (sizeof(long) == BO_LONG_SIZE) {
		lpmax = lp + nelem;
		nelem = 0;
		while (lp < lpmax) {
			l = *lp++;
			LONG_BIGENDIAN(l);
			nelem += mfil_write((CharPtr)&l, 1, BO_LONG_SIZE, mfp);
		}
		return nelem;
	}
	lpmax = lp + nelem;
	nelem = 0;
	while (lp < lpmax) {
		l = *lp++;
		LONG_BIGENDIAN(l);
		nelem += mfil_write(((CharPtr)&l)+(sizeof(long)-BO_LONG_SIZE), 1, BO_LONG_SIZE, mfp);
	}
	return nelem;
#endif
}

long PNTR _cdecl
mfil_get_long(mfp, nelem)
	MFILE	PNTR mfp;
	size_t	nelem;
{
	register size_t	off;
	long	PNTR lp0, PNTR lp, PNTR lpmax, i;
	CharPtr	cp;
	register CharPtr	ip;
	size_t	nbytes;

	nbytes = nelem * BO_LONG_SIZE;

#if BYTE_ORDER == BIG_ENDIAN
	if (sizeof(long) == BO_LONG_SIZE) {
		cp = mfil_get(mfp, nbytes);
		if (cp == NULL)
			return NULL;
		return (long *)cp;
	}
#endif

	lp0 = lp = (long PNTR)mem_malloc(nelem * sizeof(*lp));
	if (lp == NULL)
		return NULL;
	lpmax = lp + nelem;

	if (sizeof(long) == BO_LONG_SIZE) {
#ifdef YESMFILE
		if (mfil_read((CharPtr)lp, nbytes, 1, mfp) != 1) {
#else
		if (fread((CharPtr)lp, nbytes, 1, mfp) != 1) {
#endif
			mem_free(lp0);
			return NULL;
		}
		while (lp < lpmax) {
			LONG_BIGENDIAN(*lp);
			++lp;
		}
		return lp0;
	}


	ip = (CharPtr) &i;
	ip += (sizeof(i) - BO_LONG_SIZE);

#ifdef YESMFILE
	if (mfp->sp != NULL) {
		cp = mfil_get(mfp, nbytes);
		if (cp == NULL) {
			mem_free(lp0);
			return NULL;
		}
		while (lp < lpmax) {
			i = 0;
			memcpy(ip, cp, BO_LONG_SIZE);
			LONG_BIGENDIAN(i);
			*lp++ = i;
			cp += BO_LONG_SIZE;
		}
		return lp0;
	}
#endif

	while (lp < lpmax) {
		i = 0;
#ifdef YESMFILE
		if (fread(ip, BO_LONG_SIZE, 1, mfp->fp) != 1) {
#else
		if (fread(ip, BO_LONG_SIZE, 1, mfp) != 1) {
#endif
			mem_free(lp0);
			return NULL;
		}
		LONG_BIGENDIAN(i);
		*lp++ = i;
	}

	return lp0;
}

/*
	mfil_dup_long -- copy data from the memory file into the provided buffer
	of longs for 'nelem' elements
*/
size_t _cdecl
mfil_dup_long(mfp, lp, nelem)
	MFILE	PNTR mfp;
	register long	PNTR lp;
	size_t	nelem;
{
	long	i, PNTR lp0 = lp, PNTR lpmax;
	size_t	nbytes, nelem0 = nelem;
	register CharPtr	ip;

	nbytes = nelem * BO_LONG_SIZE;
	lpmax = lp + nelem;

	if (sizeof(long) == BO_LONG_SIZE) {
#ifdef YESMFILE
		if (mfil_read((CharPtr)lp, nbytes, 1, mfp) != 1)
#else
		if (fread((CharPtr)lp, nbytes, 1, mfp) != 1)
#endif
			return 0;
#if BYTE_ORDER != BIG_ENDIAN
		while (lp < lpmax) {
			LONG_BIGENDIAN(*lp);
			++lp;
		}
#endif
		return nelem0;
	}

	ip = (CharPtr) &i;
	ip += (sizeof(i) - BO_LONG_SIZE);

#ifdef YESMFILE
	if (mfp->sp != NULL) {
		register CharPtr	cp;

		cp = mfil_get(mfp, nbytes);
		if (cp == NULL)
			return NULL;
		while (lp < lpmax) {
			i = 0;
			Nlm_MemCpy(ip, cp, BO_LONG_SIZE);
			LONG_BIGENDIAN(i);
			*lp++ = i;
			cp += BO_LONG_SIZE;
		}
		return nelem0;
	}
#endif

	while (lp < lpmax) {
		i = 0;
#ifdef YESMFILE
		if (fread(ip, BO_LONG_SIZE, 1, mfp->fp) != 1)
#else
		if (fread(ip, BO_LONG_SIZE, 1, mfp) != 1)
#endif
			return lpmax - lp0;
		LONG_BIGENDIAN(i);
		*lp++ = i;
	}

	return nelem0;
}


#ifdef mfil_open
#undef mfil_open
#undef mfil_close
#undef mfil_tell
#undef mfil_eof
#undef mfil_read
#undef mfil_write
#undef mfil_seek
#endif

MFILE PNTR _cdecl
mfil_open(name, mode, flags)
	CharPtr	name, mode;
	int		flags;
{
#ifdef YESMFILE
	MFILE	mf;
	int		prot, mflags, fd;
	
	mf.fp = NULL;
	mf.shmdp = NULL;
	mf.sp = NULL;
	mf.offset = 0;
	mf.maplen = 0;
	mf.mode = *mode;

	if (strcmp(name, "-") && (flags & MFIL_OPT_SHM)) {
		mf.shmdp = shm_attfile(name, mode, FALSE);
		mf.sp = (CharPtr)shm_addr(mf.shmdp);
		if (mf.sp != NULL) {
			mf.maplen = mf.filesize = mf.shmdp->filesize;
			return (MFILE PNTR) mem_dup(&mf, sizeof(mf));
		}
	}

	mf.fp = openfile(name, mode);
	if (mf.fp == NULL)
		return NULL;

#ifdef MMAP_AVAIL
	if (strcmp(name, "-") && (flags & MFIL_OPT_MMAP)) {
		switch (*mode) {
		case 'w':
			mflags = MAP_SHARED | MAP_AUTOGROW | MAP_FILE;
			prot = PROT_WRITE;
			break;
		case 'a':
			mflags = MAP_SHARED | MAP_AUTOGROW | MAP_FILE;
			prot = PROT_READ | PROT_WRITE;
			break;
		case 'r':
		default:
			mflags = MAP_SHARED | MAP_FILE;
			prot = PROT_READ;
		}
		mf.filesize = sys_filesize(name);
		fd = fileno(mf.fp);
		if (*mode == 'r')
			mf.maplen = mf.filesize;
		else {
			mf.maplen = maplen;
			if (*mode == 'w')
				mf.filesize = 0;
			else
				mf.offset = mf.filesize;
		}
		mf.sp = mmap(NULL, mf.maplen, prot, mflags, fd, 0);
		/* Leave the file open if write-accessible */
		if (mf.sp != NULL && *mode == 'r') {
			fclose(mf.fp);
			mf.fp = NULL;
		}
	}
#endif /* MMAP_AVAIL */

	return (MFILE PNTR) mem_dup(&mf, sizeof(mf));
#else /* !YESMFILE */
	return openfile(name, mode);
#endif /* !YESMFILE */
}

int _cdecl
mfil_close(mfp)
	MFILE	PNTR mfp;
{
#ifdef YESMFILE
	if (mfp == NULL)
		return -1;
	if (mfp->sp != NULL) {
		if (mfp->shmdp != NULL)
			(void) shm_dropfile(mfp->shmdp);
#ifdef MMAP_AVAIL
		else {
			munmap(mfp->sp, mfp->filesize);
			if (mfp->fp != NULL)
				ftruncate(fileno(mfp->fp), mfp->filesize);
		}
#endif /* MMAP_AVAIL */
	}
	if (mfp->fp != NULL)
		(void) fclose(mfp->fp);
	mfp->shmdp = NULL;
	mfp->sp = NULL;
	mfp->fp = NULL;
	mem_free(mfp);
	return 0;
#else /* !YESMFILE */
	return fclose(mfp);
#endif /* !YESMFILE */
}

long int _cdecl
mfil_tell(mfp)
	MFILE	PNTR mfp;
{
#ifdef YESMFILE
	if (mfp->sp != NULL)
		return mfp->offset;
	return ftell(mfp->fp);
#else /* !YESMFILE */
	return ftell(mfp);
#endif /* !YESMFILE */
}

int _cdecl
mfil_eof(mfp)
	MFILE	PNTR mfp;
{
#ifdef YESMFILE
	if (mfp->sp != NULL) {
		if (mfp->shmdp != NULL) {
			if (mfp->offset >= mfp->filesize)
				return 1;
			return 0;
		}
		else {
			if (mfp->mode == 'r') {
				if (mfp->offset >= mfp->filesize)
					return 1;
				return 0;
			}
			else {
				if (mfp->offset >= mfp->maplen)
					return 1;
				return 0;
			}
		}
	}
	return feof(mfp->fp);
#else /* !YESMFILE */
	return feof(mfp);
#endif /* !YESMFILE */
}

/* mfil_read -- analog to ANSI C fread() */
int _cdecl
mfil_read(buf, size, nitems, mfp)
	register CharPtr	buf;
	size_t	size, nitems;
	MFILE	PNTR mfp;
{
#ifdef YESMFILE
	register size_t	off, len;
	register CharPtr	cp, cpmax;

	if (mfp->sp != NULL) {
		len = size * nitems;
		off = mfp->offset;
		if (off + len > mfp->filesize) {
			nitems = (mfp->filesize - off) / size;
			len = nitems * size;
		}
		mfp->offset += len;
		cp = mfp->sp + off;
		cpmax = cp + len;
		while (cp < cpmax) {
			*buf++ = *cp++;
		}
		return nitems;
	}
	return fread(buf, size, nitems, mfp->fp);
#else /* !YESMFILE */
	return fread(buf, size, nitems, mfp);
#endif
}

/* mfil_write -- analog to ANSI C fwrite() */
int _cdecl
mfil_write(buf, size, nitems, mfp)
	register CharPtr	buf;
	size_t	size, nitems;
	MFILE	PNTR mfp;
{
#ifdef YESMFILE
	register size_t	len;

	if (mfp->sp != NULL) {
		len = size * nitems;
		if (mfp->shmdp != NULL) {
			/* Not permitted to write past the end of a shared memory segment */
			if (mfp->offset + len > mfp->filesize) {
				nitems = (mfp->filesize - mfp->offset) / size;
				len = nitems * size;
			}
		}
		else {
			if (mfp->offset + len > mfp->maplen) {
				nitems = (mfp->maplen - mfp->offset) / size;
				len = nitems * size;
			}
		}
		Nlm_MemCpy(mfp->sp + mfp->offset, buf, len);
		mfp->offset += len;
		if (mfp->offset > mfp->filesize)
			mfp->filesize = mfp->offset;
		return nitems;
	}
	return fwrite(buf, size, nitems, mfp->fp);
#else /* !YESMFILE */
	return fwrite(buf, size, nitems, mfp);
#endif /* !YESMFILE */
}

/*
	mfil_seek() -- reposition the memory file pointer
*/
int _cdecl
mfil_seek(mfp, offset, ptrname)
	MFILE	PNTR mfp;
	long	offset;
	int		ptrname;
{
#ifdef YESMFILE
	register long	i;

	if (mfp->sp != NULL) {
		switch (ptrname) {
			case SEEK_SET: /* relative to beginning */
				if (offset < 0 || offset > mfp->filesize)
					return -1;
				mfp->offset = offset;
				break;
			case SEEK_CUR: /* relative to current position */
				i = mfp->offset + offset;
				if (i > mfp->maplen || i < 0)
					return -1;
				if (i > mfp->filesize)
					mfp->filesize = i;
				mfp->offset = i;
				break;
			case SEEK_END: /* relative to end of file */
				if (mfp->filesize < offset || offset < 0)
					return -1;
				mfp->offset = mfp->filesize - offset;
				break;
			default:
				return -1;
		}
		return 0;
	}
	return fseek(mfp->fp, offset, ptrname);
#else /* !YESMFILE */
	return fseek(mfp, offset, ptrname);
#endif /* !YESMFILE */
}
