/* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================*/
#include <ncbi.h>
#include <gishlib.h>

CharPtr _cdecl
vstr_cpy(base, maxlenp, dst, src)
	CharPtr	PNTR base, dst, src;
	size_t	PNTR maxlenp;
{
	register size_t	maxlen, remains, srclen, off;
	register CharPtr	tmpbase;

	off = dst - (tmpbase = *base);
	remains = (maxlen = *maxlenp) - off;
	srclen = strlen(src);
	if (srclen < remains) {
		Nlm_MemCpy(dst, src, srclen+1);
		return dst+srclen;
	}
	++srclen;
	if (tmpbase == NULL)
		tmpbase = mem_malloc(srclen);
	else
		tmpbase = mem_realloc(tmpbase, maxlen + (srclen - remains));
	if (tmpbase == NULL)
		return NULL;
	*base = tmpbase;
	*maxlenp = maxlen + srclen - remains;
	Nlm_MemCpy(dst = tmpbase + off, src, srclen);
	return dst+srclen-1;
}

/* Avoid calling vstr_cat() -- it will be slow! */
int _cdecl
vstr_cat(dstp, maxlenp, src)
	register CharPtr	PNTR dstp, src;
	register size_t	PNTR maxlenp;
{
	register size_t	len, dlen;
	register CharPtr	new;

	dlen= 0;
	if (*dstp != NULL)
		dlen = strlen(*dstp);
	len = strlen(src) + 1;
	if (dlen + len <= *maxlenp) {
		Nlm_MemCpy(*dstp + dlen, src, len);
		return 0;
	}
	new = mem_realloc(*dstp, dlen + len);
	if (new == NULL)
		return 1;
	Nlm_MemCpy((*dstp = new) + dlen, src, *maxlenp = len);
	return 0;
}
